#  tests for kwant-1.4.2-py37hc8548de_0 (this is a generated file);
print('===== testing package: kwant-1.4.2-py37hc8548de_0 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
import sys
import os.path

from pytest import main

import kwant

def test():
    kwant_path = os.path.dirname(os.path.abspath(kwant.__file__))
    sys.exit(main([kwant_path, "-x", "-v"]))

test()
#  --- run_test.py (end) ---

print('===== kwant-1.4.2-py37hc8548de_0 OK =====');
print("import: 'kwant'")
import kwant

print("import: 'kwant.graph'")
import kwant.graph

print("import: 'kwant.graph.tests'")
import kwant.graph.tests

print("import: 'kwant.linalg'")
import kwant.linalg

print("import: 'kwant.linalg.tests'")
import kwant.linalg.tests

print("import: 'kwant.physics'")
import kwant.physics

print("import: 'kwant.physics.tests'")
import kwant.physics.tests

print("import: 'kwant.solvers'")
import kwant.solvers

print("import: 'kwant.solvers.tests'")
import kwant.solvers.tests

print("import: 'kwant.tests'")
import kwant.tests

